/*the printhead tab takes care of everything related to the printhead.

  A few terms to help any user.
  Nozzle: An electromechanical device used to deposit a droplet of ink
  Primitive: A high current pin connected directly to a nozzle
  Address: A low current pin connected to activate a nozzle
  (both the address and primitive need to be high for a nozzle to fire)
  Pulse: A single trigger of all required primitives on a single address
  Burst: A trigger of all the required nozzles on all required address ONCE
  Sweep: A smooth, uninterupted printing motion including multiple bursts in quick succesion
*/

/*
   Todo list
   add enable to the basic functions
   check nozzle tables, some holes in the tests and prints

   !Test splitting the burst in 2 and 3 halves to see if it improves the burst
*/

//defines
#define printhead_address_delay 1
#define printhead_warmup_limit 50000


const uint8_t printhead_enable = 2;
const uint8_t printhead_address_reset = 8; //D3
const uint8_t printhead_address_next = 9; //C3
const uint8_t printhead_shift_clock = 17; //B1
const uint8_t printhead_shift_data = 23; //C2
const uint8_t printhead_shift_latch = 16; //BO
const uint8_t printhead_primitive_clock = 15; //C0
const uint8_t printhead_primitive_clear = 14; //D1
const uint8_t printhead_check_nozzle = 21; //D6
const uint8_t printhead_10XR = A6; //D20
const uint8_t printhead_TSR = A8; //22

//nozzle tables
const uint8_t printhead_nozzle_table_address[300] = //the table that returns the address for a given nozzle
{
  15, 12, 7, 4, 12, 21, 4, 5, 21, 14,
  5, 3, 14, 20, 3, 13, 20, 18, 13, 0,
  18, 11, 0, 8, 11, 16, 8, 1, 16, 19,
  1, 2, 19, 10, 2, 9, 21, 17, 9, 6,
  17, 15, 6, 16, 15, 12, 7, 4, 12, 21,
  4, 5, 21, 14, 5, 3, 14, 20, 3, 13,
  20, 18, 13, 0, 18, 11, 0, 8, 11, 16,
  8, 1, 16, 19, 1, 2, 19, 10, 2, 9,
  10, 17, 9, 6, 17, 15, 6, 7, 15, 12,
  7, 4, 12, 21, 4, 5, 21, 14, 5, 3,
  14, 20, 3, 13, 20, 18, 13, 0, 18, 11,
  0, 8, 11, 16, 8, 1, 16, 19, 1, 2,
  19, 10, 2, 9, 10, 17, 9, 6, 17, 15,
  6, 7, 15, 12, 7, 4, 12, 21, 4, 5,
  21, 14, 5, 3, 14, 20, 3, 13, 20, 18,
  13, 0, 18, 11, 0, 8, 11, 16, 8, 1,
  16, 19, 1, 2, 19, 10, 2, 9, 10, 17,
  15, 6, 7, 15, 6, 7, 15, 12, 7, 4,
  12, 21, 4, 5, 18, 14, 5, 3, 14, 20,
  3, 13, 20, 18, 13, 0, 18, 11, 0, 8,
  11, 16, 8, 1, 16, 19, 1, 2, 19, 10,
  2, 9, 10, 17, 9, 6, 17, 15, 6, 7,
  15, 12, 7, 4, 12, 21, 4, 5, 21, 14,
  5, 3, 14, 20, 3, 13, 20, 18, 13, 0,
  18, 11, 0, 9, 11, 16, 8, 1, 16, 19,
  1, 2, 19, 10, 2, 9, 10, 17, 9, 6,
  17, 15, 6, 7, 15, 12, 7, 4, 12, 21,
  4, 5, 21, 14, 5, 3, 14, 20, 3, 13,
  20, 18, 13, 0, 18, 11, 0, 8, 11, 16,
  8, 1, 16, 19, 1, 2, 19, 10, 2, 19
};

const uint8_t printhead_nozzle_table_primitive[300] = //the table that returns the primitive for a given nozzle
{
  4, 10, 4, 10, 4, 10, 4, 10, 4, 10,
  4, 10, 4, 10, 4, 10, 4, 10, 4, 10,
  4, 10, 4, 10, 4, 10, 4, 10, 4, 10,
  4, 10, 4, 10, 4, 10, 4, 10, 4, 10,
  3, 9, 3, 9, 3, 9, 3, 9, 3, 9,
  3, 9, 3, 9, 3, 9, 3, 9, 3, 9,
  3, 9, 3, 9, 3, 9, 3, 9, 3, 9,
  3, 9, 3, 9, 3, 9, 3, 9, 3, 9,
  3, 9, 3, 9, 1, 8, 1, 8, 1, 8,
  1, 8, 1, 8, 1, 8, 1, 8, 1, 8,
  1, 8, 1, 8, 1, 8, 1, 8, 1, 8,
  1, 8, 1, 8, 1, 8, 1, 8, 1, 8,
  1, 8, 1, 8, 1, 8, 1, 8, 0, 14,
  0, 14, 0, 14, 0, 14, 0, 14, 0, 14,
  0, 14, 0, 14, 0, 14, 0, 14, 0, 14,
  0, 14, 0, 14, 0, 14, 0, 14, 0, 14,
  0, 14, 0, 14, 0, 14, 0, 14, 0, 14,
  0, 14, 5, 13, 5, 13, 5, 13, 5, 13,
  5, 13, 5, 13, 5, 13, 5, 13, 5, 13,
  5, 13, 5, 13, 5, 13, 5, 13, 5, 13,
  5, 13, 5, 13, 5, 13, 5, 13, 5, 13,
  5, 13, 5, 13, 5, 13, 6, 12, 6, 12,
  6, 12, 6, 12, 6, 12, 6, 12, 6, 12,
  6, 12, 6, 12, 6, 12, 6, 12, 6, 12,
  6, 12, 6, 12, 6, 12, 6, 12, 6, 12,
  6, 12, 6, 12, 6, 12, 6, 12, 6, 12,
  2, 11, 2, 11, 2, 11, 2, 11, 2, 11,
  2, 11, 2, 11, 2, 11, 2, 11, 2, 11,
  2, 11, 2, 11, 2, 11, 2, 11, 2, 11,
  2, 11, 2, 11, 2, 11, 2, 11, 2, 11

};

//primitive nozzle table is split in 2 halves because of buffer efficiency reasons. Printhead has 2 rows of nozzles spaced 4ish millimeters)
//the split allows for 2 different positions in the buffer to be read efficiently (hopefully)
//left blank for now, maybe it will be generated with the table generation, maybe it won't, who knows
/*
  const uint8_t printhead_nozzle_table_primitive_B1[150] = //the table that returns the primitive for a given nozzle (gives the first byte)
  {

  };

  const uint8_t printhead_nozzle_table_primitive_B2[150] = //the table that returns the primitive for a given nozzle (gives the second byte)
  {

  };
*/


int16_t printhead_nozzle_table_nozzle[22][16]; //the table that returns the nozzle for a given address and primitive (returns -1 if the nozzle does not exist) (table is generated by the firmware) (a contains 16 for shift register reasons)
uint8_t printhead_address_percentage_table[22]; //how many primitives are in an address (at most 14)
uint8_t printhead_primitive_percentage_table[14]; //how many addresses are in a primitive (at most 22);

uint8_t printhead_enabled;



void PrintheadStart() { //sets all pins for inkjet and sets all the pins to the default
  //set up the pins
  pinMode(printhead_enable, OUTPUT); //set the enable pin to an output
  pinMode(printhead_address_reset, OUTPUT);
  pinMode(printhead_address_next, OUTPUT);
  pinMode(printhead_shift_clock, OUTPUT);
  pinMode(printhead_shift_data, OUTPUT);
  pinMode(printhead_shift_latch, OUTPUT);
  pinMode(printhead_primitive_clock, OUTPUT);
  pinMode(printhead_primitive_clear, OUTPUT);
  pinMode(printhead_check_nozzle, INPUT);

  //set the pins to default
  digitalWrite(printhead_enable, 0); //set the enable pin to the default (low)
  digitalWrite(printhead_address_reset, 0);
  digitalWrite(printhead_address_next, 0);
  digitalWrite(printhead_shift_clock, 0);
  digitalWrite(printhead_shift_data, 0);
  digitalWrite(printhead_shift_latch, 0);
  digitalWrite(printhead_primitive_clock, 0);
  digitalWrite(printhead_primitive_clear, 0);

  PrintheadDisable(); //disable the printhead

  //set the address to 0
  PrintheadAddressReset();

  //generate tables
  PrintheadGenerateReverseTable(); //generate reverse nozzle table
  PrintheadGeneratePercentageTable(); //generate percentage tables

}






//Inkjet functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
//printhead burst. Prints an entire burst (all addresses, single run) from the data passed into the function as a word array of 22 (uint16_t temp_print[22]).
//the argument gets printed according to the nozzle tables, starting at the first address and ending at the 22nd.
//this function uses the rawest functions to make the speed as high as possible
void PrintheadInkjetBurstW(uint16_t temp_print[]) {
  //Serial.println("\n new line:");
  uint16_t temp_blank_array = 0; //#create the blank array for the end
  const uint8_t temp_start_points[4] = {2, 4, 6, 8}; //#where the address reset turns on, off and address next turns on, off
  const uint8_t temp_shift_points[3] = {5, 9, 11}; //where the primitive clear is triggered, address next turns on and off (4 is around 1.7us, 5 is around )
  const uint16_t temp_overlay_array[2] = {43690, 21845}; //{37449, 18724, 9362}; //{43690, 21845}; //10 recuring and 01 recuring, triggering halve)
  uint8_t temp_point = 0; //the next point to look for in the points arrays
  uint8_t temp_overlay_pos = 0; //which of the overlay positions is used
  uint8_t temp_printing_state; //if the data in the register needs to be printed or not
  uint8_t temp_last_pulse = 0;
  uint16_t temp_trigger_array, temp_trigger_compensated; //the array that is to be printed

  //set primitive clock and clear
  PrintheadPrimitiveClearLow();
  PrintheadPrimitiveClockLow();

  //reset and next
  PrintheadAddressReset();
  PrintheadAddressNext();

  temp_printing_state = 0;
  for (int8_t a = 0; a < 23; a++) { //loop through all the data and pulse. Starts at -1 to latch the first data before triggering
    if (a == 22) { //if 22, hollow array latch data and pulse
      temp_trigger_array = 0;
      temp_last_pulse = 1; //mark last pulse
    }
    else { //if 0-21, latch data and pulse (if required)
      temp_trigger_array = temp_print[a];
    }

    for (uint8_t x = 0; x < 2; x++) { //for 2 times
      if (temp_last_pulse == 2) break; //break from loop if last pulse
      temp_point = 0; //reset state value

      temp_trigger_compensated = temp_trigger_array & temp_overlay_array[x];  //overlay value
      if (temp_printing_state == 1) { //if system can latch data
        PrintheadPrimitiveClearHigh();
        noInterrupts(); //disable interrupts <---------------------!
        PrintheadPrimitiveClockHigh(); //when the first address is shifted in, latch in the data to the primitives
      }
      else { //first position clock needs to be ignored
        temp_printing_state = 1; 
      }
      //pulse printhead
      PrintheadShiftLatchLow();
      for (uint8_t s = 0; s < 16; s++) {
        //set the next bit of the shift register
        if (temp_trigger_compensated >> s & 1) { //if the currently read bit is high
          PrintheadShiftDataHigh();
        }
        else {
          PrintheadShiftDataLow();
        }
        PrintheadShiftClockHigh();
        PrintheadShiftClockLow(); //clock data

        //see if the change condition are met
        if (s == temp_shift_points[temp_point]) { //check if another pin needs to change
          if (temp_point == 0) { //clear the primitive
            PrintheadPrimitiveClearLow();
            PrintheadPrimitiveClockLow();
            interrupts(); //enable interrupts <-------------------------------------------!
          }
          else if (temp_point == 1) { //address next high
            PrintheadAddressNextHigh();
          }
          else if (temp_point == 2) { //address next low
            PrintheadAddressNextLow();
            temp_point = 0; //reset to 0 to prevent wrong access to array (one ++ will not break it)
          }
          temp_point++; //go to the next point
        }
      }
      PrintheadShiftLatchHigh(); //latch the data
      if (temp_last_pulse == 1) temp_last_pulse == 2; //if last pulse was done, mark end 
    }
  }//*/


  /*PrintheadShiftLatchLow(); //set the latch to low
  for (uint8_t s = 0; s < 16; s++) { //shift the first pulse in
    //set data
    if (temp_print[0] >> s & 1) { //if the currently read bit is high
      PrintheadShiftDataHigh();
    }
    else {
      PrintheadShiftDataLow();
    }
    PrintheadShiftClockHigh();
    PrintheadShiftClockLow(); //clock data

    //see if the change condition are met
    if (s == temp_start_points[temp_point]) {
      if (temp_point == 0) { //address reset high
        PrintheadAddressResetHigh();
      }
      else if (temp_point == 1) { //address reset low
        PrintheadAddressResetLow();

      }
      else if (temp_point == 2) { //address next high
        PrintheadAddressNextHigh();
      }
      else if (temp_point == 3) { //address next low
        PrintheadAddressNextLow();
        temp_point = 0; //reset to 0 to prevent wrong access to array (one ++ will not break it)
      }
      temp_point++; //go to the next point
    }
  }
  PrintheadShiftLatchHigh(); //latch the data

  //go into the loop, where the shifting starts during the pulse
  for (uint8_t a = 1; a < 22; a++) {
    temp_point = 0; //reset state value
    PrintheadPrimitiveClearHigh();
    noInterrupts(); //disable interrupts <---------------------!
    PrintheadPrimitiveClockHigh(); //when the first address is shifted in, latch in the data to the primitives
    //Serial.println(temp_print[a]);
    PrintheadShiftLatchLow();
    for (uint8_t s = 0; s < 16; s++) {
      //set the next bit of the shift register
      if (temp_print[a] >> s & 1) { //if the currently read bit is high
        PrintheadShiftDataHigh();
      }
      else {
        PrintheadShiftDataLow();
      }
      PrintheadShiftClockHigh();
      PrintheadShiftClockLow(); //clock data

      //see if the change condition are met
      if (s == temp_shift_points[temp_point]) { //check if another pin needs to change
        if (temp_point == 0) { //clear the primitive
          PrintheadPrimitiveClearLow();
          PrintheadPrimitiveClockLow();
          interrupts(); //enable interrupts <-------------------------------------------!
        }
        else if (temp_point == 1) { //address next high
          PrintheadAddressNextHigh();
        }
        else if (temp_point == 2) { //address next low
          PrintheadAddressNextLow();
          temp_point = 0; //reset to 0 to prevent wrong access to array (one ++ will not break it)
        }
        temp_point++; //go to the next point
      }
    }
    PrintheadShiftLatchHigh(); //latch the data
  }


  //(the last pulse will actually be done with blank data to preserve the timing) -----
  temp_point = 0; //reset state value
  uint16_t temp_hollow_array = 0;
  PrintheadPrimitiveClearHigh();
  PrintheadPrimitiveClockHigh(); //when the first address is shifted in, latch in the data to the primitives
  PrintheadShiftLatchLow();
  for (uint8_t s = 0; s < 16; s++) {
    //set the next bit of the shift register
    if (temp_hollow_array >> s & 1) { //if the currently read bit is high
      PrintheadShiftDataHigh();
    }
    else {
      PrintheadShiftDataLow();
    }
    PrintheadShiftClockHigh();
    PrintheadShiftClockLow(); //clock data

    //see if the change condition are met
    if (s == temp_shift_points[temp_point]) { //check if another pin needs to change
      if (temp_point == 0) { //clear the primitive
        PrintheadPrimitiveClearLow();
        PrintheadPrimitiveClockLow();
      }
      else if (temp_point == 1) { //address next high
        PrintheadAddressNextHigh();
      }
      else if (temp_point == 2) { //address next low
        PrintheadAddressNextLow();
        temp_point = 0; //reset to 0 to prevent wrong access to array (one ++ will not break it)
      }
      temp_point++; //go to the next point
    }
  }
  PrintheadShiftLatchHigh(); //latch the data*/
}

void PrintheadInkjetBurst2B(uint8_t temp_print_1[], uint8_t temp_print_2[]) {

}



void PrintheadInkjetPulse2B(uint8_t temp_input_1, uint8_t temp_input_2) { //pulses the current address with the data input (slower and inaccurate, but simple code)

  PrintheadPrimitiveClearLow();
  PrintheadPrimitiveClockLow();
  PrintheadPrimitiveClearHigh();
  PrintheadShiftSet2B(temp_input_1, temp_input_2);
  PrintheadPrimitiveClockHigh();

  delayMicroseconds(2);

  PrintheadPrimitiveClearLow();
  PrintheadPrimitiveClockLow();
}

void PrintheadInkjetPulseW(uint16_t temp_input) {
  PrintheadPrimitiveClearLow();
  PrintheadPrimitiveClockLow();
  PrintheadPrimitiveClearHigh();
  PrintheadShiftSetW(temp_input);
  PrintheadPrimitiveClockHigh();

  delayMicroseconds(2);

  PrintheadPrimitiveClearLow();
  PrintheadPrimitiveClockLow();
}


void PrintheadPreheat(uint32_t temp_pulses) {
  //enable printhead
  uint8_t temp_enable_history = PrintheadEnabled();
  if (temp_enable_history == 0) {
    PrintheadSetEnable(1);
    //Serial.println("enabled");
  }

  temp_pulses = constrain(temp_pulses, 0, printhead_warmup_limit); //limit the max number of pulses

  //heating time around 750ns
  PrintheadShiftSetW(65535); //set shift register

  //repeat for the number of pulses
  for (uint32_t p = 0; p < temp_pulses; p++) {
    PrintheadAddressReset();

    for (uint8_t a = 0; a < 22; a++) {
      PrintheadAddressNext(); //go to next address
      delayMicroseconds(printhead_address_delay);

      PrintheadPrimitiveClearHigh();
      PrintheadPrimitiveClockHigh();

      //delay
      __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns
      __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns
      __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns

      PrintheadPrimitiveClearLow();
      PrintheadPrimitiveClockLow();
      delayMicroseconds(printhead_address_delay);
    }
  }
  PrintheadAddressReset();
  //disable is printhead was disabled
  if (temp_enable_history == 0) PrintheadSetEnable(0);
}


//warning, temporary test function. Don't get attached to it too much. It will not be here in this for much longer
//it is used to print a single line, one nozzle at a time, odd or even to test if the nozzle tables are in order.
// see this:   v    , that says temporary, didn't I tell you?
void PrintheadTempTriggerRow() {
  PrintheadEnable();
  delay(1);

  for (uint16_t n = 1; n < 300; n += 2) { //walk through all odd or even nozzles
    //generate print values
    uint8_t temp_address = printhead_nozzle_table_address[n];
    uint8_t temp_primitive_val = printhead_nozzle_table_primitive[n];
    uint16_t temp_primitive_trigger_val = 0;
    bitWrite(temp_primitive_trigger_val, temp_primitive_val, 1);
    uint8_t temp_prim_low = lowByte(temp_primitive_trigger_val);
    uint8_t temp_prim_high = highByte(temp_primitive_trigger_val);

    for (uint8_t r = 0; r < 10; r++) {
      PrintheadAddressReset();
      for (uint8_t a = 0; a < 22; a++) {
        PrintheadAddressNext();
        if (a == temp_address) {
          PrintheadInkjetPulse2B(temp_prim_low, temp_prim_high);
        }
        else {
          delayMicroseconds(5);
        }
      }
    }
    delay(10);
  }
  PrintheadDisable();
}


//takes an array of 300 int8_t and fills them with the pulses to charge of each nozzle
//to do: A single set of nozzles (around 7) is not found. They appear to be in random locations.
//This indicates a mistake in the nozzle reverse table, but where is not known.
void PrintheadTestNozzles(int8_t* temp_nozzle_state) {
  //prepare values
  int16_t temp_nozzle;
  uint16_t temp_primtive_value;
  const int8_t temp_pulse_quantity = 20; //the number of pulses to charge the capacitor per nozzle
  uint8_t temp_test_pulses;

  //reset the values
  for (uint16_t n = 0; n < 300; n++) {
    temp_nozzle_state[n] = -1;
  }

  PrintheadAddressReset(); //reset the address

  for (uint8_t a = 0; a < 22; a++) { //loop through all the addresses
    PrintheadAddressNext(); //go to the next address
    for (uint8_t p = 0; p < 16; p++) { //loop through all the primitives
      PrintheadEnable(); //start draining the capacitor
      temp_nozzle = PrintheadGetNozzle(a, p); //get the nozzle
      temp_primtive_value = 0; //reset the shift register value
      bitWrite(temp_primtive_value, p, 1); //create the shift register value
      PrintheadShiftSetW(temp_primtive_value); //set the value to the shift register
      delayMicroseconds(40); //added delay to drain the capacitor
      PrintheadDisable(); //close the printhead, stop draining


      if (temp_nozzle >= 0) {//start testing the printhead if the nozzle exists
        PrintheadPrimitiveClearLow();
        PrintheadPrimitiveClockLow();
        temp_test_pulses = 0;
        while (temp_test_pulses < temp_pulse_quantity) { //test until the number of pulses is reached or the test conditions are true
          //pulse the head
          PrintheadPrimitiveClearHigh();
          PrintheadPrimitiveClockHigh();

          __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns
          __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns
          __asm__("nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t""nop\n\t"); //18 NOP's, ca. 200ns

          PrintheadPrimitiveClearLow();
          PrintheadPrimitiveClockLow();

          //see if the check pin has become low
          delayMicroseconds(50); //delay
          if (digitalRead(printhead_check_nozzle) == 0) { //if yes, break
            break;
          }
          //if no, continue

          temp_test_pulses++; //add one to the number of pulses done
        }

        //write to value
        //Serial.print(temp_nozzle); Serial.print(", "); Serial.println(temp_test_pulses);
        temp_nozzle_state[temp_nozzle] = temp_test_pulses;
      }
    }
  }
  PrintheadAddressReset(); //reset the address back to 0 to prevent temperature errors
}


//595 shift register functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
void PrintheadShiftSet2B(uint8_t temp_input_1, uint8_t temp_input_2) { //sets the shift register simply to the values given in the input (2 bytes)
  PrintheadShiftLatchLow(); //latch low

  //loop through all the available data
  //byte 1
  for (uint8_t b = 0; b < 8; b++) {
    //set data
    if (temp_input_1 >> b & 1) { //if the currently read bit is high
      PrintheadShiftDataHigh();
    }
    else {
      PrintheadShiftDataLow();
    }
    //pulse clock
    PrintheadShiftClockHigh();
    PrintheadShiftClockLow();
  }

  //byte 2
  for (uint8_t b = 0; b < 8; b++) {
    //set data
    //set data
    if (temp_input_2 >> b & 1) { //if the currently read bit is high
      PrintheadShiftDataHigh();
    }
    else {
      PrintheadShiftDataLow();
    }
    //pulse clock
    PrintheadShiftClockHigh();
    PrintheadShiftClockLow();
  }


  PrintheadShiftLatchHigh(); //latch high
}

void PrintheadShiftSetW(uint16_t temp_input) { //sets the shift register simply to the values given in the input (1 word)
  PrintheadShiftLatchLow(); //latch low

  //loop through all the available data
  for (uint8_t w = 0; w < 16; w++) {
    //set data
    if (temp_input >> w & 1) { //if the currently read bit is high
      PrintheadShiftDataHigh();
    }
    else {
      PrintheadShiftDataLow();
    }
    //pulse clock
    PrintheadShiftClockHigh();
    PrintheadShiftClockLow();
  }


  PrintheadShiftLatchHigh(); //latch high
}

void PrintheadShiftClockHigh() { //make the Shigt Clock High
  GPIOB_PDOR = GPIOB_PDOR | B00000010; //make Shift Clock pin high
  //digitalWrite(printhead_shift_clock, 1); //make Shift Clock pin high
}

void PrintheadShiftClockLow() { //make the Shift Clock Low
  GPIOB_PDOR = GPIOB_PDOR & B11111101; //make Shift Clock pin low
  //digitalWrite(printhead_shift_clock, 0);  //make Shift Clock pin low
}

void PrintheadShiftDataHigh() { //make the Shift Data High
  GPIOC_PDOR = GPIOC_PDOR | B00000100; //make Shift Data pin high
  //digitalWrite(printhead_shift_data, 1); //make Shift Data pin high
}

void PrintheadShiftDataLow() { //make the Shift Data Low
  GPIOC_PDOR = GPIOC_PDOR & B11111011; //make Shift Data pin low
  //digitalWrite(printhead_shift_data, 0);  //make Shift Data pin low
}

void PrintheadShiftLatchHigh() { //make the Shift Latch High
  GPIOB_PDOR = GPIOB_PDOR | B00000001; //make Shift Latch pin high
  //digitalWrite(printhead_shift_latch, 1); //make Shift Latch pin high
}

void PrintheadShiftLatchLow() { //make the Shift Latch Low
  GPIOB_PDOR = GPIOB_PDOR & B11111110; //make Shift Latch pin low
  //digitalWrite(printhead_shift_latch, 0);  //make Shift Latch pin low
}


//primitive functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
void PrintheadPrimitiveClockHigh() { //make the Primitive Clock High
  GPIOC_PDOR = GPIOC_PDOR | B00000001; //make Primitive Clock pin high
  //digitalWrite(printhead_primitive_clock, 1); //make Primitive Clock pin high
}

void PrintheadPrimitiveClockLow() { //make the Primitive Clock Low
  GPIOC_PDOR = GPIOC_PDOR & B11111110; //make Primitive Clock pin low
  //digitalWrite(printhead_primitive_clock, 0);  //make Primitive Clock pin low
}

void PrintheadPrimitiveClearHigh() { //make the Primitive Clear High
  GPIOD_PDOR = GPIOD_PDOR | B00000010; //make Primitive Clear pin high
  //digitalWrite(printhead_primitive_clear, 1); //make Primitive Clear pin high
}

void PrintheadPrimitiveClearLow() { //make the Primitive Clear Low
  GPIOD_PDOR = GPIOD_PDOR & B11111101; //make Primitive Clear pin low
  //digitalWrite(printhead_primitive_clear, 0);  //make Primitive Clear pin low
}


//address functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
void PrintheadAddressNext() { //step to the next address
  PrintheadAddressNextHigh(); //make address next pin high
  delayMicroseconds(printhead_address_delay); //wait pin register as high
  PrintheadAddressNextLow(); //make address next pin low
  delayMicroseconds(printhead_address_delay); //wait pin register as low
}

void PrintheadAddressNextHigh() { //make address next pin high
  GPIOC_PDOR = GPIOC_PDOR | B00001000; //make address next pin high
  //digitalWrite(printhead_address_next, 1); //make address next pin high
}

void PrintheadAddressNextLow() { //make addres next pin low
  GPIOC_PDOR = GPIOC_PDOR & B11110111; //make address next pin low
  //digitalWrite(printhead_address_next, 0);  //make address next pin low
}

void PrintheadAddressReset() { //reset the address to -1
  PrintheadAddressResetHigh(); //make address reset pin high
  delayMicroseconds(printhead_address_delay); //wait pin register as high
  PrintheadAddressResetLow(); //make address reset pin low
  delayMicroseconds(printhead_address_delay); //wait pin register as low
}

void PrintheadAddressResetHigh() { //make address reset pin high
  GPIOD_PDOR = GPIOD_PDOR | B00001000; //make address reset pin low
  //digitalWrite(printhead_address_reset, 1); //make address reset pin low
}

void PrintheadAddressResetLow() { //make address reset pin low
  GPIOD_PDOR = GPIOD_PDOR & B11110111; //make address reset pin low
  //digitalWrite(printhead_address_reset, 0); //make address reset pin low
}

//enable functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
void PrintheadEnable() { //enable the printhead mosfet, connecting the printhead ground to the common ground
  digitalWrite(printhead_enable, 1);
  printhead_enabled = 1; //set enabled state
}

void PrintheadDisable() { //disable the printhead mosfet, severing the connection between the printhead ground to the common ground
  digitalWrite(printhead_enable, 0);
  printhead_enabled = 0; //set enabled state
}

void PrintheadSetEnable(uint8_t temp_input) { //enable or disable the printhead using an input value 1 or 0. All others are ignored
  if (temp_input > 1) return; //return from function if input does not match
  if (temp_input == 1) {
    PrintheadEnable();
  }
  else {
    PrintheadDisable();
  }
}

uint8_t PrintheadEnabled() { //returns the state of the printhead enabled flag
  return printhead_enabled;
}



//analog functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
//read raw 10x
//returns -1 if head is not enabled
int16_t PrintheadGet10XRaw() {
  if (printhead_enabled == 0) { //check if the printhead is enabled
    return -1;
  }
  analogReadResolution(13); //set resolution to 13 bits
  int32_t temp_analog = analogRead(printhead_10XR);
  analogReadResolution(10); //set resolution back to 10 bits
  return temp_analog; //return the raw analog value
}


//read raw TSR
int16_t PrintheadGetTSRRaw() {
  if (printhead_enabled == 0) { //check if the printhead is enabled
    return -1;
  }
  analogReadResolution(13); //set resolution to 13 bits
  int32_t temp_analog = analogRead(printhead_TSR);
  analogReadResolution(10); //set resolution back to 10 bits
  return temp_analog; //return the raw analog value
}

//read corrected Temperature
//returns the temperature in 0.1. 200 is 20.0 (Celcius or Fahrenheit)
//mode 1 is celcius, mode 2 is fahrenheit
//-1 is printhead not enabled, -2 is values out of bounds
//might not be that accurate, but seems accurate enough on hp45. HP15 is wildly out
int16_t PrintheadGetTemperature(uint8_t temp_mode) {
  if (temp_mode == 1 || temp_mode == 2) {
    if (printhead_enabled == 0) { //check if the printhead is enabled
      return -1;
    }
    int16_t temp_10x = PrintheadGet10XRaw();
    int16_t temp_tsr = PrintheadGetTSRRaw();
    const float temp_vin = 3.3;
    const float temp_r1 = 330.0;

    //Vout = (R2 / (R2 + R1)) * Vin, R1 is 330ohm
    //R2 = ((Vout x R1)/(Vin-Vout))

    //calculate in celcius
    float temp_f_calc;
    //get the 10x voltage
    temp_f_calc = float(temp_10x);
    temp_f_calc /= 8192; //get the fraction
    temp_f_calc *= temp_vin; //get the voltage
    float temp_10x_res = ((temp_f_calc * temp_r1) / (temp_vin - temp_f_calc)); //get the 10x resistance
    //Serial.print("10X resistor: "); Serial.println(temp_10x_res);

    //get the tsr voltage
    temp_f_calc = float(temp_tsr);
    temp_f_calc /= 8192; //get the fraction
    temp_f_calc *= temp_vin; //get the voltage
    float temp_tsr_res = ((temp_f_calc * temp_r1) / (temp_vin - temp_f_calc)); //get the tsr resistance
    //Serial.print("RSR resistor: "); Serial.println(temp_tsr_res);

    //check both to be reasonable values
    if (temp_10x_res < 150.0 || temp_10x_res > 500.0) return -2;
    if (temp_tsr_res < 150.0 || temp_tsr_res > 500.0) return -2;

    //get the TSR - 10X
    //at 10 ohms, the temperature is 20C, for every 1.1 ohms the temperature rises 1 degree
    //T = 1.1R+10
    temp_f_calc = temp_tsr_res - temp_10x_res;
    float temp_temp = 1.1 * temp_f_calc + 10.0;
    temp_temp *= 10.0; //multiply by ten to add the decimal value
    //Serial.print("Temperature: "); Serial.println(temp_temp);
    if (temp_mode == 1) return long(temp_temp); //if celcius
    else { //convert to fahrenheit (lol)
      temp_temp = (temp_temp * (9.0 / 5.0)) + 320.0;
      return long(temp_temp);
    }
  }
}



//table functions ------------------------------------------------------------------------------------------------------------------------------------------------------------
//fetch requests from the existing tables
//return address value from nozzle
uint8_t PrintheadGetAddress(uint16_t temp_nozzle) {
  temp_nozzle = constrain(temp_nozzle, 0, 299);
  return printhead_nozzle_table_address[temp_nozzle];
}

//return primitive value from nozzle
uint8_t PrintheadGetPrimitive(uint16_t temp_nozzle) {
  temp_nozzle = constrain(temp_nozzle, 0, 299);
  return printhead_nozzle_table_primitive[temp_nozzle];
}

//return nozzle from address and primitive
int16_t PrintheadGetNozzle(uint8_t temp_address, uint8_t temp_primitive) {
  temp_address = constrain(temp_address, 0, 21);
  temp_primitive = constrain(temp_primitive, 0, 15);
  return printhead_nozzle_table_nozzle[temp_address][temp_primitive];
}

//return the percentage (most possible primitives in an address) of a given address
int16_t PrintheadGetAddressPercentage(uint8_t temp_input) {
  temp_input = constrain(temp_input, 0, 21);
  return printhead_address_percentage_table[temp_input];
}

//return the percentage (most possible addresses in a primitive) of a given primitive
int16_t PrintheadGetPrimitivePercentage(uint8_t temp_input) {
  temp_input = constrain(temp_input, 0, 13);
  return printhead_primitive_percentage_table[temp_input];
}


//generation functions to create tables from the 2 raw tables
//generate reverse table used to lookup a primitive and address and returns a nozzle. table (not this function) returns a -1 if the address and primitive pair have no nozzle
void PrintheadGenerateReverseTable() {
  uint8_t temp_address;
  uint8_t temp_primitive;
  for (uint8_t a = 0; a < 22; a++) {
    for (uint8_t p = 0; p < 16; p++) {
      printhead_nozzle_table_nozzle[a][p] = -1; //first fill all the positions with -1. everything not mentioned remains -1
    }
  }

  for (uint16_t n = 0; n < 300; n++) { //walk through all the nozzles
    temp_address = PrintheadGetAddress(n); //get the address of the value
    temp_primitive = PrintheadGetPrimitive(n); //get the primitive
    printhead_nozzle_table_nozzle[temp_address][temp_primitive] = n; //fill the reverse value [a][p] with the nozzle
  }
}


//generate percentage values for address and primitives. Used to indicate the percentage of operational nozzles on a given primitive and address
void PrintheadGeneratePercentageTable() {

  //reset the arrays to 0
  for (uint8_t a = 0; a < 22; a++) {
    printhead_address_percentage_table[a] = 0;
  }
  for (uint8_t p = 0; p < 16; p++) {
    printhead_primitive_percentage_table[p] = 0;
  }

  //loop through all the nozzles, add 1 to each position that matches
  for (uint16_t n = 0; n < 300; n++) {
    printhead_address_percentage_table[PrintheadGetPrimitive(n)] ++; //primitive is added to the address table
    printhead_primitive_percentage_table[PrintheadGetAddress(n)] ++; //address is added to the primitive table
  }
}


//the generate tables function is used to generate, well, new tables. It takes the raw printhead tables, an conversion table that takes the raw values and alters them
//and the prints the new table over serial. All you need to do is correct the conversion table, run the function, and copy, paste the new table in the place of the old.
void PrintheadGenerateNewRawTables() {
  //raw printhead locations
  const uint8_t temp_raw_nozzle_table_address[300] = //the table that returns the address for a given nozzle
  {
    0, 8, 15, 1, 8, 16, 1, 9, 16, 2,
    9, 17, 2, 10, 17, 3, 10, 18, 3, 11,
    18, 4, 11, 19, 4, 12, 19, 5, 12, 20,
    5, 13, 20, 6, 13, 21, 16, 14, 21, 7,
    14, 0, 7, 12, 0, 8, 15, 1, 8, 16,
    1, 9, 16, 2, 9, 17, 2, 10, 17, 3,
    10, 18, 3, 11, 18, 4, 11, 19, 4, 12,
    19, 5, 12, 20, 5, 13, 20, 6, 13, 21,
    6, 14, 21, 7, 14, 0, 7, 15, 0, 8,
    15, 1, 8, 16, 1, 9, 16, 2, 9, 17,
    2, 10, 17, 3, 10, 18, 3, 11, 18, 4,
    11, 19, 4, 12, 19, 5, 12, 20, 5, 13,
    20, 6, 13, 21, 6, 14, 21, 7, 14, 0,
    7, 15, 0, 8, 15, 1, 8, 16, 1, 9,
    16, 2, 9, 17, 2, 10, 17, 3, 10, 18,
    3, 11, 18, 4, 11, 19, 4, 12, 19, 5,
    12, 20, 5, 13, 20, 6, 13, 21, 6, 14,
    0, 7, 15, 0, 7, 15, 0, 8, 15, 1,
    8, 16, 1, 9, 18, 2, 9, 17, 2, 10,
    17, 3, 10, 18, 3, 11, 18, 4, 11, 19,
    4, 12, 19, 5, 12, 20, 5, 13, 20, 6,
    13, 21, 6, 14, 21, 7, 14, 0, 7, 15,
    0, 8, 15, 1, 8, 16, 1, 9, 16, 2,
    9, 17, 2, 10, 17, 3, 10, 18, 3, 11,
    18, 4, 11, 21, 4, 12, 19, 5, 12, 20,
    5, 13, 20, 6, 13, 21, 6, 14, 21, 7,
    14, 0, 7, 15, 0, 8, 15, 1, 8, 16,
    1, 9, 16, 2, 9, 17, 2, 10, 17, 3,
    10, 18, 3, 11, 18, 4, 11, 19, 4, 12,
    19, 5, 12, 20, 5, 13, 20, 6, 13, 20
  };

  const uint8_t temp_raw_nozzle_table_primitive[300] =
  {
    0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
    0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
    0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
    0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
    2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
    2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
    2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
    2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
    2, 3, 2, 3, 4, 5, 4, 5, 4, 5,
    4, 5, 4, 5, 4, 5, 4, 5, 4, 5,
    4, 5, 4, 5, 4, 5, 4, 5, 4, 5,
    4, 5, 4, 5, 4, 5, 4, 5, 4, 5,
    4, 5, 4, 5, 4, 5, 4, 5, 6, 7,
    6, 7, 6, 7, 6, 7, 6, 7, 6, 7,
    6, 7, 6, 7, 6, 7, 6, 7, 6, 7,
    6, 7, 6, 7, 6, 7, 6, 7, 6, 7,
    6, 7, 6, 7, 6, 7, 6, 7, 6, 7,
    6, 7, 8, 9, 8, 9, 8, 9, 8, 9,
    8, 9, 8, 9, 8, 9, 8, 9, 8, 9,
    8, 9, 8, 9, 8, 9, 8, 9, 8, 9,
    8, 9, 8, 9, 8, 9, 8, 9, 8, 9,
    8, 9, 8, 9, 8, 9, 10, 11, 10, 11,
    10, 11, 10, 11, 10, 11, 10, 11, 10, 11,
    10, 11, 10, 11, 10, 11, 10, 11, 10, 11,
    10, 11, 10, 11, 10, 11, 10, 11, 10, 11,
    10, 11, 10, 11, 10, 11, 10, 11, 10, 11,
    12, 13, 12, 13, 12, 13, 12, 13, 12, 13,
    12, 13, 12, 13, 12, 13, 12, 13, 12, 13,
    12, 13, 12, 13, 12, 13, 12, 13, 12, 13,
    12, 13, 12, 13, 12, 13, 12, 13, 12, 13
  };

  //conversion values
  int8_t temp_primitive_conversion[14] = {4, 10, 3, 9, 1, 8, 0, 14, 5, 13, 6, 12, 2, 11};
  uint8_t temp_address_conversion[22] = {15, 4, 14, 13, 11, 1, 10, 6, 12, 5, 20, 0, 16, 2, 17, 7, 21, 3, 18, 8, 19, 9};

  //primitive
  Serial.println("New Primitive Array");
  for (uint16_t n = 0; n < 300; n++) { //walk through all values
    uint8_t temp_val = temp_raw_nozzle_table_primitive[n]; //take the current value
    temp_val = temp_primitive_conversion[temp_val]; //convert it using the conversion table
    Serial.print(temp_val); //print new value
    if (n != 299) { //don't print the last comma
      Serial.print(","); //comma
    }
    if (n % 10 == 9) { //new line after 10 positions
      Serial.println(""); //<-- new line
    }
  } //<- curly bracket
  // <- Blank space :)
  //address
  Serial.println("");
  Serial.println("New Address Array");
  for (uint16_t n = 0; n < 300; n++) { //walk through all values
    uint8_t temp_val = temp_raw_nozzle_table_address[n]; //take the current value
    temp_val = temp_address_conversion[temp_val]; //convert it using the conversion table
    Serial.print(temp_val); //print new value
    if (n != 299) { //don't print the last comma
      Serial.print(","); //comma
    }
    if (n % 10 == 9) { //new line after 10 positions
      Serial.println(""); //<-- new line
    }
  }
}


