//printing tab holds the more advance inkjet modes. It is more focused around multiple bursts and inkjet decoding from nozzle information
//perhaps in the future some of the printhead command also should be moved here, but I am not really sure of that


/*broad functions:
   decode from nozzle raw
   decode from toggle
   decode from a/p
*/

/*
   Todo list

   Direct print modes:
   decoding from toggle
   decoding from A/P

   separate text buffer
   text to buffer interface

   Decoding to buffer:
   buffer interface

   Trigger modes:
   Timed mode without trigger (asap)
   Timed mode with trigger
   Direct trigger mode
   Position mode, encoder
   Position mode, time dictated, without trigger
   Position mode, time dictated, with trigger

  future: barcode support
  Far future, QR code support
*/

uint8_t printing_odd_even_mode; //0 for all, 1 for odd rows, 2 for even rows
uint16_t printing_dpi; //possible values: 600, 300, 200, 150, 100, 75 (odd or even requires halve DPI)

uint8_t printing_mode_nozzle_start; //at which nozzle the decoding starts
uint8_t printing_mode_nozzle_step; //how many nozzles the decoding steps after each value
uint16_t printing_mode_nozzle_repeat; //how many times the read value is repeated before the next one is fetched

//the printing array is the data for printing a single burst, in 2B format (buffer reasons)
//if in odd or even mode, the buffer is used [0][0], [1][0], [0][1], [1][1] to use all space
uint8_t printing_array[22][2];


//Start printing
void PrintingStart() { //sets all values right
  printing_dpi = 600;
  printing_odd_even_mode = 0;
  PrintingRewrite();
  PrintheadSetEnable(0); //disable head, (again)
}



//print the current value in the array
void PrintingPrintArray() {
  //enable printhead
  uint8_t temp_enable_history = PrintheadEnabled();
  if (temp_enable_history == 0) {
    PrintheadSetEnable(1);
    //Serial.println("enabled");
  }

  //temporary decode from 2B to W
  uint16_t temp_array[22];
  for (uint8_t a = 0; a < 22; a++) {
    //Serial.print("array: "); Serial.println(printing_array[a][1]);
    temp_array[a] = printing_array[a][1];
    temp_array[a] = temp_array[a] << 8;
    temp_array[a] = temp_array[a] | printing_array[a][0];
    //Serial.print("array: "); Serial.println(temp_array[a]);
  }
  PrintheadInkjetBurstW(temp_array); //print array

  //disable is printhead was disabled
  if (temp_enable_history == 0) PrintheadSetEnable(0);
}

//decode raw mode (from base 64 raw to the printing array)
void PrintingDecodeRaw(uint8_t temp_input_values[], uint16_t temp_array_size) {
  //Serial.println("Decoding base64 raw");
  PrintingResetArray(); //reset the array
  uint16_t temp_nozzle = printing_mode_nozzle_start; //current nozzle variable
  uint16_t temp_repeat = 0; //the current position in the repeat
  uint8_t temp_read_position = 0; //goes from bit 0 to bit 5
  //Serial.println(printing_mode_nozzle_start);

  uint16_t temp_array_read = 0; //where in the input array the reading is
  uint16_t temp_array_end = temp_array_size; //where the reading stops

  uint8_t temp_address, temp_primitive, temp_primitive_converted;

  while (1) { //start looping through the array
    if (temp_input_values[temp_array_read] >> temp_read_position & 1) { //read value, if 1
      temp_address = PrintheadGetAddress(temp_nozzle); //look up address
      temp_primitive = PrintheadGetPrimitive(temp_nozzle); //look up primitive
      /*Serial.print("nozzle: "); Serial.print(temp_nozzle);
        Serial.print(", reading: "); Serial.print(temp_array_read);
        Serial.print(", bit: "); Serial.print(temp_read_position);
        Serial.print(", Address: "); Serial.print(temp_address);
        Serial.print(", Primitive: "); Serial.println(temp_primitive);//*/
      //get right halve of primitive (replace with %8 and /8 to simplify later)
      if (temp_primitive < 8) { //lower halve
        temp_primitive_converted = 0; //reset converted so only the bit shows
        bitWrite(temp_primitive_converted, temp_primitive, 1); //write the right bit in the primitive
        printing_array[temp_address][0] =  printing_array[temp_address][0] | temp_primitive_converted; //overlay value with array
        //Serial.print(", P array 0 "); Serial.println(temp_primitive_converted);
      }
      else { //upper halve
        temp_primitive_converted = 0;
        temp_primitive -= 8; //subtract 8 to make value a byte
        bitWrite(temp_primitive_converted, temp_primitive, 1); //write the right bit in the primitive
        printing_array[temp_address][1] = printing_array[temp_address][1] | temp_primitive_converted; //overlay value with array
        //Serial.print(", P array 1 "); Serial.println(temp_primitive_converted);
      }//*/
    }

    //next read value conditions (quite a few really, because of all the DPI and odd even stuff)
    temp_nozzle += printing_mode_nozzle_step; //step nozzle to next position
    if (temp_nozzle >= 300) { //if end of printhead, stop reading
      //Serial.println("End of nozzles");
      break;
    }
    //check overflow of: repeat value -> read position -> end of array
    temp_repeat ++; //add one to repeat
    if (temp_repeat >= printing_mode_nozzle_repeat) { //if repeat overflows
      temp_repeat = 0; //reset repeat
      temp_read_position++; //go to the next read position
      if (temp_read_position > 5) { //if read position moves past 5, go to next value
        temp_read_position = 0;
        temp_array_read++; //go to next value in the array
        if (temp_array_read >= temp_array_end) { //if end of array is found (value is first unfilled value)
          //Serial.println("End of array");
          break;
        }
      }
    }
  } //end of while 1 loop
}


//decode toggle mode (from base 64 toggle to printing array)
void PrintingToggleRaw(uint8_t temp_input_values[], uint16_t temp_array_size) {

}


//decode a/p mode (not really that much decoding to do, pass values to printing array if valid)



//reset printing array
void PrintingResetArray() {
  for (uint8_t a = 0; a < 22; a++) {
    printing_array[a][0] = 0;
    printing_array[a][1] = 0;
  }
}


//set and modify printing variables
//recalculate start, step and repeat
void PrintingRewrite() {
  //set start based on odd/even
  if (printing_odd_even_mode == 0 || printing_odd_even_mode == 2) printing_mode_nozzle_start = 0; //odd/even or even, start at 0
  else printing_mode_nozzle_start = 1; //start at one

  //set step based on odd even
  if (printing_odd_even_mode == 0) printing_mode_nozzle_step = 1; //odd/even mode, 1 nozzle per step
  else printing_mode_nozzle_step = 2; //odd or even mode, 2 nozzle per step

  //modify DPI based on odd/even, DPI can not be over 300 in odd or even mode

  if (printing_odd_even_mode != 0) { //if odd/even is not active
    if (printing_dpi > 300) {
      printing_dpi = 300;
    }
  }

  //set repeat based on odd/even
  printing_mode_nozzle_repeat = 600 / printing_mode_nozzle_step; //amount of nozzles per inch. divide by step to compensate odd/even
  printing_mode_nozzle_repeat /= printing_dpi; //divide by DPI to get number of repeats. repeat will be every whole number available from 1. At 1 DPI, repeats will be 300/600
}

//set DPI mode
void PrintingSetDPI(uint16_t temp_input) {
  temp_input = constrain(temp_input, 1, 600); //constrain values (if you give more as an input, its not my problem) (0 would cause problems)
  printing_dpi = temp_input;
  PrintingRewrite(); //rewirte values
}

//set Odd/even mode
void PrintingSetOddEven(uint8_t temp_input) {
  temp_input = constrain(temp_input, 0, 2); //constrain values (if you give more as an input, its not my problem)
  printing_odd_even_mode = temp_input;
  PrintingRewrite(); //rewrite values
}




