/*
   Serial functions hold all the functions that can only ben handled by the serial interface. This include functions that are Serial specific returns and serial related information
   Functions that are more broad are handled by other tabs
*/

/*
   Todo list
*/

#define serial_function_test_threshold 5 //the threshold after which a nozzle is considered broken


//tests the printhead and returns the results over serial.
//0-9 is reserved for computer code, 10-19 is reserved for human readable
//ending in 0, send all
//ending in 1, send nozzles of nozzles working
//ending in 2, send nozzle table
//ending in 5, send address percentages
//ending in 6, send primitive percentages
void SerialFunctionTestPrinthead(uint8_t temp_mode) {
  //test printhead
  int8_t temp_nozzle_state[300]; //create nozzle table
  PrintheadTestNozzles(temp_nozzle_state); //test the printhead

  //check if a test can be done with the mode
  //-----

  //return basic answer
  Serial.print("TST ");
  Serial.print(temp_mode);
  Serial.println(":");


  uint8_t temp_mode_start = temp_mode / 10; //create first digit
  //Serial.print("mode start: "); Serial.println(temp_mode_start);
  uint8_t temp_mode_end = temp_mode % 10; //create the final digit
  //Serial.print("mode end: "); Serial.println(temp_mode_end);
  if (temp_mode_end == 0 || temp_mode_end == 1) { //calculate and send amount of nozzles working
    //Calculate
    uint16_t temp_working_nozzles = 0;
    for (uint16_t n = 0; n < 300; n++) {
      if (temp_nozzle_state[n] < serial_function_test_threshold && temp_nozzle_state[n] != -1) { //if the nozzle is past the threshold
        temp_working_nozzles++; //mark one more nozzle as working
      }
    }

    //print result
    if (temp_mode_start == 0) { //computer readable
      Serial.print("NZW:");
      Serial.println(temp_working_nozzles);
      //Serial.println("/300");
    }
    else if (temp_mode_start == 1) { //human readable
      Serial.print("Nozzles working: ");
      Serial.print(temp_working_nozzles);
      Serial.println("/300");
    }
  }
  if (temp_mode_end == 0 || temp_mode_end == 2) {   //calculate and send whole nozzle table
    //print result
    if (temp_mode_start == 0) { //computer readable
      Serial.print("NTB:"); //print header
      //Calculate
      for (uint16_t n = 0; n < 300; n++) { //loop through all nozzles
        //if (n % 10 == 0) Serial.println(""); //go to new line every tenth nozzle
        //maybe send data in different base, base 32 makes sense, base 64 is already present
        if (temp_nozzle_state[n] < serial_function_test_threshold && temp_nozzle_state[n] != -1) {
          Serial.print(1);
        }
        else {
          Serial.print(0);
        }
      }
      Serial.println("");
    }
    else if (temp_mode_start == 1) { //human readable
      Serial.print("Nozzle Table:"); //print header
      //Calculate
      for (uint16_t n = 0; n < 300; n++) { //loop through all nozzles
        if (n % 10 == 0) {
          Serial.println(""); //go to new line every tenth nozzle
          Serial.print(n); //print the nozzle at the start, every tenth nozzle
          if (n < 10) Serial.print(" "); //add a space if the number is smaller than 10
          if (n < 100) Serial.print(" "); //add a space if the number is smaller than 100
          Serial.print(":");
        }
        if (temp_nozzle_state[n] < serial_function_test_threshold && temp_nozzle_state[n] != -1) {
          Serial.print(1);
        }
        else {
          Serial.print(0);
        }
        Serial.print(", ");
      }
      Serial.println("");
    }
  }
  if (temp_mode_end == 0 || temp_mode_end == 5) {   //calculate and send address fraction
    //Calculate
    uint8_t temp_address_array[22]; //create array to write to
    for (uint8_t a = 0; a < 22; a++) {
      temp_address_array[a] = 0; //reset the array
    }

    //see how many working primitives there are in each address
    for (uint16_t n = 0; n < 300; n++) { //loop through all the nozzles
      if (temp_nozzle_state[n] < serial_function_test_threshold && temp_nozzle_state[n] != -1) {
        //if a nozzle is working, add it to the right address
        temp_address_array[PrintheadGetAddress(n)] ++;
      }
    }
    //print result
    //you will get primitives and addresses mixed up here. That is normal.
    if (temp_mode_start == 0) { //computer readable
      Serial.print("ATB:");
      for (uint8_t a = 0; a < 22; a++) {
        Serial.print(temp_address_array[a]);
        Serial.print("/");
        Serial.print(PrintheadGetPrimitivePercentage(a));
        Serial.print(",");
      }
      Serial.println("");
    }
    else if (temp_mode_start == 1) { //human readable
      Serial.println("Addresses Working:");
      for (uint8_t a = 0; a < 22; a++) {
        Serial.print("A");
        Serial.print(a);
        if (a < 10) Serial.print(" ");
        Serial.print (": ");
        Serial.print(temp_address_array[a]);
        Serial.print("/");
        Serial.println(PrintheadGetPrimitivePercentage(a));
      }
    }
  }
  if (temp_mode_end == 0 || temp_mode_end == 6) {   //calculate and send primitive fraction
    //Calculate
    uint8_t temp_primitive_array[16]; //create array to write to
    for (uint8_t p = 0; p < 16; p++) {
      temp_primitive_array[p] = 0; //reset the array
    }

    //see how many working addresses there are in each primitive
    for (uint16_t n = 0; n < 300; n++) { //loop through all the nozzles
      if (temp_nozzle_state[n] < serial_function_test_threshold && temp_nozzle_state[n] != -1) {
        //if a nozzle is working, add it to the right address
        temp_primitive_array[PrintheadGetPrimitive(n)] ++;
      }
    }

    //print result
    //you will get primitives and addresses mixed up here. That is normal.
    if (temp_mode_start == 0) { //computer readable
      Serial.print("PTB:");
      for (uint8_t p = 0; p < 16; p++) {
        Serial.print(temp_primitive_array[p]);
        Serial.print("/");
        Serial.print(PrintheadGetAddressPercentage(p));
        Serial.print(",");
      }
      Serial.println("");
    }
    else if (temp_mode_start == 1) { //human readable
      Serial.println("Primitives Working:");
      for (uint8_t p = 0; p < 16; p++) {
        Serial.print("P");
        Serial.print(p);
        if (p < 10) Serial.print(" ");
        Serial.print (": ");
        Serial.print(temp_primitive_array[p]);
        Serial.print("/");
        Serial.println(PrintheadGetAddressPercentage(p));
      }
    }
  }

  //end the transmission
  if (temp_mode_start == 0) { //computer readable
    Serial.println("End");
  }
  else if (temp_mode_start == 1) { //human readable
    Serial.println("End of Test");
  }
}

void SerialFunctionGetTemperature(uint8_t temp_mode) {
  //Get temperature
  //check enable, if enabled, leave enabled, if disabled, temporarily enable
  uint8_t temp_enable_history = PrintheadEnabled();
  if (temp_enable_history == 0) PrintheadSetEnable(1); delay(1);
  int16_t temp_temperature;
  if (temp_mode == 1) {
    temp_temperature = PrintheadGetTemperature(1); //get celcius
    if (temp_enable_history == 0) PrintheadSetEnable(0);
  }
  else if (temp_mode == 2 || temp_mode == 101) { //heh, lol
    temp_temperature = PrintheadGetTemperature(2); //get fahrenheit
    if (temp_enable_history == 0) PrintheadSetEnable(0);
  }
  else { //If no good op value, return error
    Serial.print("GTP ");
    Serial.print(temp_mode);
    Serial.println(":Error");
    if (temp_enable_history == 0) PrintheadSetEnable(0);
    return;
  }

  //Send return code over serial
  Serial.print("GTP ");
  Serial.print(temp_mode);
  Serial.print(":");
  //Send temperature
  Serial.println(temp_temperature);

}

//address next serial command
void SerialFunctionAddressNext(uint8_t temp_input) {
  if (temp_input == 0) { //off
    PrintheadAddressNextLow();
  }
  if (temp_input == 1) { //on
    PrintheadAddressNextHigh();
  }
  if (temp_input == 10) { //single pulse
    PrintheadAddressNext();
  }
}

//address reset serial command
void SerialFunctionAddressReset(uint8_t temp_input) {
  if (temp_input == 0) { //off
    PrintheadAddressResetLow();
  }
  if (temp_input == 1) { //on
    PrintheadAddressResetHigh();
  }
  if (temp_input == 10) { //single pulse
    PrintheadAddressReset();
  }
}

//primitive clock serial command
void SerialFunctionPrimitiveClock(uint8_t temp_input) {
  if (temp_input == 0) { //off
    PrintheadPrimitiveClockLow();
  }
  if (temp_input == 1) { //on
    PrintheadPrimitiveClockHigh();
  }
}

//primitive clear serial command
void SerialFunctionPrimitiveClear(uint8_t temp_input) {
  if (temp_input == 0) { //off
    PrintheadPrimitiveClearLow();
  }
  if (temp_input == 1) { //on
    PrintheadPrimitiveClearHigh();
  }
}

//shift register command
void SerialFunctionSetShift(uint16_t temp_input) {
  PrintheadShiftSetW(temp_input);
}

//return tables ------------
//return address table value
int16_t SerialFunctionGetAddressTable(uint16_t temp_input) {
  if (temp_input > 299) return -1; //if the value is incorrect
  Serial.print("NTA ");
  Serial.print(temp_input);
  Serial.print(":");
  Serial.println(PrintheadGetAddress(temp_input));
}

//return primitive table value
int16_t SerialFunctionGetPrimitiveTable(uint16_t temp_input) {
  if (temp_input > 299) return -1; //if the value is incorrect
  Serial.print("NTP ");
  Serial.print(temp_input);
  Serial.print(":");
  Serial.println(PrintheadGetPrimitive(temp_input));
}


